/*
$Id:$

NT7534_I2C LCD library! (for GTK-281)

Copyright (C) 2016 Guy WEILER www.weigu.lu

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

some of this code was written by <cstone@pobox.com> originally and
some of this code was written by Limor Fried, Adafruit Industries
it is in the public domain.
*/

#ifndef __CL_I2C_LCD_H__
#define __CL_I2C_LCD_H__

#include "Arduino.h"

#define MCP23017_addr 0x20
#define MCP23017_PortA 0x12  // data
#define MCP23017_PortB 0x13  // res, bl, e, rw, di, 3PB

#define swap(a, b) { uint8_t t = a; a = b; b = t; }

#define BLACK 1
#define WHITE 0

#define LCDWIDTH                 128
#define LCDHEIGHT                64

// Commands see data sheet NT7534_I2C
#define CMD_DISPLAY_OFF          0xAE
#define CMD_DISPLAY_ON           0xAF
#define CMD_SET_DISP_START_LINE  0x40           //From 0x40 to 0x7F
#define CMD_SET_PAGE             0xB0
#define CMD_SET_COLUMN_UPPER     0x10
#define CMD_SET_COLUMN_LOWER     0x00
#define CMD_SET_ADC_NORMAL       0xA0
#define CMD_SET_ADC_REVERSE      0xA1
#define CMD_SET_DISP_NORMAL      0xA6
#define CMD_SET_DISP_REVERSE     0xA7
#define CMD_SET_ALLPTS_NORMAL    0xA4
#define CMD_SET_ALLPTS_ON        0xA5
#define CMD_SET_BIAS_9           0xA2
#define CMD_SET_BIAS_7           0xA3
#define CMD_SET_COM_NORMAL       0xC0
#define CMD_SET_COM_REVERSE      0xC8
#define CMD_SET_POWER_CONTROL    0x28
#define CMD_SET_RESISTOR_RATIO   0x20
#define CMD_SET_VOLUME_FIRST     0x81
#define CMD_SET_VOLUME_SECOND    0x00
#define CMD_SET_STATIC_OFF       0xAC
#define CMD_SET_STATIC_ON        0xAD
#define CMD_SET_STATIC_REG       0x00
#define CMD_SET_BOOSTER_FIRST    0xF8
#define CMD_SET_BOOSTER_234         0
#define CMD_SET_BOOSTER_5           1
#define CMD_SET_BOOSTER_6           3
#define CMD_RMW                  0xE0
#define CMD_RMW_CLEAR            0xEE
#define CMD_INTERNAL_RESET       0xE2
#define CMD_NOP                  0xE3
#define CMD_TEST                 0xF0

class NT7534_I2C {
 public:
  // 8 Data Pins on MCP23017 PORTA
  NT7534_I2C(uint8_t rst, uint8_t bl, uint8_t en, uint8_t rw, uint8_t di):
        rst(rst), bl(bl), di(di), rw (rw), en(en)  {}

  void pinPortB(uint8_t c, boolean l);
  void setPort(uint8_t port, uint8_t c);
  uint8_t getPort(uint8_t port);
  void command(uint8_t c);
  void datawrite(uint8_t c);
  uint8_t dataread();
  void init(void);
  void begin(uint8_t contrast);
  void setbrightness(uint8_t val);
  void clear();
  void setall(void);
  void setpixel(uint8_t x, uint8_t y, uint8_t color);
  bool getpixel(uint8_t x, uint8_t y);
  void setbyte(uint8_t x, uint8_t p, uint8_t b);
  uint8_t getbyte(uint8_t x, uint8_t p);
  void drawline(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, uint8_t color);
  void fillcircle(uint8_t x0, uint8_t y0, uint8_t r, uint8_t color);
  void drawcircle(uint8_t x0, uint8_t y0, uint8_t r, uint8_t color);
  void drawrect(uint8_t x, uint8_t y, uint8_t w, uint8_t h, uint8_t color);
  void fillrect(uint8_t x, uint8_t y, uint8_t w, uint8_t h, uint8_t color);
  void displaylogo(uint8_t nr);
  void printchar(uint8_t x, uint8_t line, char c);
  void clearchar(uint8_t x, uint8_t line, char c);
  void drawstring(uint8_t x, uint8_t line, const char *str);
  void clearstring(uint8_t x, uint8_t line, const char *str);
  void drawbitmap(uint8_t x, uint8_t y, const uint8_t *bitmap, uint8_t w,
                  uint8_t h, uint8_t color);

  void begin2(int x, int y, uint8_t contrast);
  void setCursor(uint8_t col, uint8_t row);
  void print(uint8_t x, uint8_t line, const char *str);
  void print(uint8_t line, const char *str);
  void print(const char *str);
  void print(int i);
  void write(char c);

  private:
  uint8_t rst, bl, en, rw, di;
  int gx, gy;

};

#endif
