/*
 * This file is part of CL Open EVSE.

 * CL Open EVSE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.

 * CL Open EVSE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with CL Open EVSE; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "../../../src/clever.h"

#define TOP ((F_CPU / 2000000) * 1000) // for 1KHz (=1000us period)

void CPilot::Init() {
  // set up Timer for phase & frequency correct PWM 1/8000 resolution
  TCCR1A = 0;  // set up Control Register A
  ICR1 = TOP;
  // WGM13 -> select P&F mode CS10 -> prescaler = 1
  TCCR1B = _BV(WGM13) | _BV(CS10);
  DDRD |= _BV(PORTD5);
  TCCR1A |= _BV(COM1A1);
  SetState(PILOT_STATE_P12); // turns the pilot on 12V steady state
}
// no PWM pilot signal - steady state
// PILOT_STATE_P12 = steady +12V (EVSE_STATE_A - VEHICLE NOT CONNECTED)
// PILOT_STATE_N12 = steady -12V (EVSE_STATE_F - FAULT)
void CPilot::SetState(PILOT_STATE state) {
  AutoCriticalSection asc;
  if (state == PILOT_STATE_P12) { OCR1A = TOP; }
  else { OCR1A = 0; }
  m_State = state;
}

// set EVSE current capacity in Amperes duty cycle
int CPilot::SetPWM(int amps) {
  unsigned cnt;
  // amps = (duty cycle %) X 0.6
  if ((amps >= 6) && (amps <= 51)) { cnt = amps * (TOP/60); }
  else if ((amps > 51) && (amps <= 80)) {
    cnt = (amps * (TOP/250)) + (64*(TOP/100)); //amps=(duty cycle % - 64) X 2.5
  }
  else { return 1; }
  OCR1A = cnt;
  m_State = PILOT_STATE_PWM;
  return 0;
}
