/* NT7534_I2C LCD library! (for GTK-281)

Copyright (C) 2016 Guy WEILER www.weigu.lu

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

some of this code was written by <cstone@pobox.com> originally and
some of this code was written by Limor Fried, Adafruit Industries
it is in the public domain.  */

#ifdef __AVR__
#include <avr/pgmspace.h>
#include <util/delay.h>
#endif

#include <stdlib.h>
#include "cl_i2c_nt7534_lcd_128x64.h"
// create logo with gimp (128x64) and save as *.xbm. Convert with xbm2NT7534_I2C.py
#include "logo1.h" //creative-lab logo

#include "Wire.h" // I2C

uint8_t is_reversed = 0;


// a handy reference to where the pages are on the screen
const uint8_t pagemap[] =  { 3, 2, 1, 0, 7, 6, 5, 4 };

// a 5x7 font table
const extern uint8_t PROGMEM font[];

void NT7534_I2C::pinPortB(uint8_t c, boolean l) {
  uint8_t PortB_MCP23017 = 0;
  Wire.beginTransmission(MCP23017_addr);
  Wire.write(MCP23017_PortB);             // PortB
  Wire.endTransmission();
  Wire.requestFrom(MCP23017_addr, 1);     // get PortB
  PortB_MCP23017 = Wire.read();
  if (l) {
    PortB_MCP23017 |= (1 << c); }  // HIGH
  else {
    PortB_MCP23017 &= ~(1 << c); } // LOW
  Wire.beginTransmission(MCP23017_addr);
  Wire.write(MCP23017_PortB);
  Wire.write(PortB_MCP23017);
  Wire.endTransmission(); }

void NT7534_I2C::setPort(uint8_t port, uint8_t c) {
  Wire.beginTransmission(MCP23017_addr);
  Wire.write(port);  // Port address
  Wire.write(c);     // data
  Wire.endTransmission(); }

uint8_t NT7534_I2C::getPort(uint8_t port) {
  uint8_t data;
  Wire.beginTransmission(MCP23017_addr);
  Wire.write(port);      // Port address
  Wire.endTransmission();
  Wire.requestFrom(MCP23017_addr, 1);     // get Port
  return( Wire.read()); }

void NT7534_I2C::command(uint8_t c) {
  pinPortB(di,LOW);	// command
  setPort(MCP23017_PortA, c);
  pinPortB(rw, LOW);	// write
  pinPortB(en, HIGH);       // rising edge
  pinPortB(en, LOW);}       // needed!

void NT7534_I2C::datawrite(uint8_t c) {
  pinPortB(di, HIGH);	// data
  setPort(MCP23017_PortA, c);
  pinPortB(rw, LOW);	// write
  pinPortB(en, HIGH);
  pinPortB(en, LOW);}        //needed!

uint8_t NT7534_I2C::dataread() {
  Wire.beginTransmission(MCP23017_addr); // PortA = Input
  Wire.write(0x00);             // IODIRA register (Bank = 0)
  Wire.write(0xFF);             // set all of port A to input
  Wire.endTransmission();
  pinPortB(di, HIGH);	// data
  pinPortB(rw, HIGH);	// read
  pinPortB(en, LOW);
  pinPortB(en, HIGH);
  //uint8_t c = PINB;
  uint8_t c = getPort(MCP23017_PortA);
  pinPortB(en, LOW);
  Wire.beginTransmission(MCP23017_addr);
  Wire.write(0x00);             // IODIRA register (Bank = 0)
  Wire.write(0x00);             // set all of port A to outputs
  Wire.endTransmission();
  return c;}

void NT7534_I2C::init(void) {
  Wire.begin();                 // wake up I2C bus
  Wire.setClock(400000L);       // 400kHz
  Wire.beginTransmission(MCP23017_addr);
  Wire.write(0x00);             // IODIRA register (Bank = 0)
  Wire.write(0x00);             // set all of port A to outputs
  Wire.endTransmission();
  Wire.beginTransmission(MCP23017_addr);
  Wire.write(0x01);             // IODIRB register (Bank = 0)
  Wire.write(0x07);             // 5 out, 3 in
  Wire.endTransmission();       // res, bl, e, rw, di, 3pb
  pinPortB(bl, HIGH);
  pinPortB(en, LOW);
  pinPortB(rst, LOW); // hardware reset the display
  _delay_ms(100);
  pinPortB(rst, HIGH);
  _delay_ms(100);
  command(CMD_SET_BIAS_7);               // LCD bias select
  command(CMD_SET_ADC_NORMAL);           // ADC select
  command(CMD_SET_COM_NORMAL);           // SHL select
  command(CMD_SET_DISP_START_LINE);      // Initial display line
  command(CMD_SET_POWER_CONTROL | 0x4);  // turn on voltage converter
  _delay_ms(50);                                // (VC=1, VR=0, VF=0)
  command(CMD_SET_POWER_CONTROL | 0x6);  // turn on voltage regulator
  _delay_ms(50);                                // (VC=1, VR=1, VF=0)
  command(CMD_SET_POWER_CONTROL | 0x7);  // turn on voltage follower
  _delay_ms(10);                                // (VC=1, VR=1, VF=1)
  command(CMD_SET_RESISTOR_RATIO | 0x6);} // set lcd operating voltage

// Para:val (5 bit) from 0x01 (small) to 0x3F (large)
void NT7534_I2C::setbrightness(uint8_t val) {
  command(CMD_SET_VOLUME_FIRST);
  command(CMD_SET_VOLUME_SECOND | (val & 0x3f));}

void NT7534_I2C::begin(uint8_t contrast) {
  init();
  command(CMD_DISPLAY_ON);
  command(CMD_SET_ALLPTS_NORMAL);
  setbrightness(contrast);}

void NT7534_I2C::begin2(int x, int y, uint8_t contrast) {
  gx = x*6;
  gy = y;
  init();
  command(CMD_DISPLAY_ON);
  command(CMD_SET_ALLPTS_NORMAL);
  setbrightness(contrast);}


void NT7534_I2C::clear(void) {
  uint8_t p, c;
  for(p = 0; p < 8; p++) {
    command(CMD_SET_PAGE | pagemap[p]);
    c=0;      // start at the beginning of the row
    command(CMD_SET_COLUMN_LOWER | ((c+1) & 0x0f));
    command(CMD_SET_COLUMN_UPPER | (((c+1) >> 4) & 0x0f));
    command(CMD_RMW);
    for(c = 0; c <= LCDWIDTH; c++) {
      datawrite(0x00);    }
    command(CMD_RMW_CLEAR);}}

void NT7534_I2C::setall(void) {
  uint8_t p, c;
  for(p = 0; p < 8; p++) {
    command(CMD_SET_PAGE | pagemap[p]);
    c=0;      // start at the beginning of the row
    command(CMD_SET_COLUMN_LOWER | ((c+1) & 0x0f));
    command(CMD_SET_COLUMN_UPPER | (((c+1) >> 4) & 0x0f));
    command(CMD_RMW);
    for(c = 0; c <= LCDWIDTH; c++) {
      datawrite(0xFF);}
    command(CMD_RMW_CLEAR);}}

void NT7534_I2C::setpixel(uint8_t x, uint8_t y, uint8_t color) {
  if ((x > LCDWIDTH) || (y > LCDHEIGHT)) return;
  uint8_t c = getbyte(x, y/8);
  command(CMD_SET_PAGE | pagemap[y/8]);
  command(CMD_SET_COLUMN_LOWER | ((x+1) & 0xf));
  command(CMD_SET_COLUMN_UPPER | (((x+1) >> 4) & 0xf));
  if (color) datawrite(c|(0x80 >> (y%8)));
  else datawrite(c & ~(0x80 >> (y%8)));}

void NT7534_I2C::setbyte(uint8_t x, uint8_t p, uint8_t b) {
  command(CMD_SET_PAGE | pagemap[p]);
  command(CMD_SET_COLUMN_LOWER | ((x+1) & 0xf));
  command(CMD_SET_COLUMN_UPPER | (((x+1) >> 4) & 0xf));
  datawrite(b);}

bool NT7534_I2C::getpixel(uint8_t x, uint8_t y) {
  uint8_t c;
  command(CMD_SET_PAGE | pagemap[y/8]);
  command(CMD_SET_COLUMN_LOWER | ((x+1) & 0xf));
  command(CMD_SET_COLUMN_UPPER | (((x+1) >> 4) & 0xf));
  c = dataread(); //dummy read see datasheet
  c = dataread();
  if ((c & ((0x80 >> (y%8))))==0) return 0;
  else return 1;}

uint8_t NT7534_I2C::getbyte(uint8_t x, uint8_t p) {
  uint8_t c;
  command(CMD_SET_PAGE | pagemap[p]);
  command(CMD_SET_COLUMN_LOWER | ((x+1) & 0xf));
  command(CMD_SET_COLUMN_UPPER | (((x+1) >> 4) & 0xf));
  c = dataread(); //dummy read see datasheet
  c = dataread();
  return c;}

void NT7534_I2C::displaylogo(uint8_t nr) {
  uint8_t p, c;
  for(p = 0; p <= 7; p++) {
    command(CMD_SET_PAGE | p);
    c=0;     // start at the beginning of the row
    command(CMD_SET_COLUMN_LOWER | (((c+1)) & 0x0f));
    command(CMD_SET_COLUMN_UPPER | ((((c+1)) >> 4) & 0x0f));
    command(CMD_RMW);
    for(c = 0; c <= LCDWIDTH; c++) {
      if (nr == 1) datawrite(pgm_read_byte(&logo1[((LCDWIDTH)*pagemap[p])+c]));
#ifdef __LOGO2_H__
      if (nr == 2) datawrite(pgm_read_byte(&logo2[((LCDWIDTH)*pagemap[p])+c]));
#endif
#ifdef __LOGO3_H__
      if (nr == 3) datawrite(pgm_read_byte(&logo3[((LCDWIDTH)*pagemap[p])+c]));
#endif
    }
    command(CMD_RMW_CLEAR);}}


void NT7534_I2C::drawline(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1,
		      uint8_t color) {  // bresenham's algorithm - thx wikpedia
  uint8_t steep = abs(y1 - y0) > abs(x1 - x0);
  if (steep) {
    swap(x0, y0);
    swap(x1, y1);}
  if (x0 > x1) {
    swap(x0, x1);
    swap(y0, y1);}
  uint8_t dx, dy;
  dx = x1 - x0;
  dy = abs(y1 - y0);
  int8_t err = dx / 2;
  int8_t ystep;
  if (y0 < y1) ystep = 1;
  else ystep = -1;
  for (; x0<=x1; x0++) {
    if (steep) setpixel(y0, x0, color);
    else setpixel(x0, y0, color);
    err -= dy;
    if (err < 0) {
      y0 += ystep;
      err += dx;}}}

void NT7534_I2C::drawrect(uint8_t x, uint8_t y, uint8_t w, uint8_t h,
		      uint8_t color) {
  for (uint8_t i=x; i<x+w; i++) {
    setpixel(i, y, color);
    setpixel(i, y+h-1, color);  }
  for (uint8_t i=y; i<y+h; i++) {
    setpixel(x, i, color);
    setpixel(x+w-1, i, color);  }}

void NT7534_I2C::fillrect(uint8_t x, uint8_t y, uint8_t w, uint8_t h,
		      uint8_t color) {
  for (uint8_t i=x; i<x+w; i++) {
    for (uint8_t j=y; j<y+h; j++) setpixel(i, j, color);}}

void NT7534_I2C::drawcircle(uint8_t x0, uint8_t y0, uint8_t r, uint8_t color) {
  int8_t f = 1 - r;
  int8_t ddF_x = 1;
  int8_t ddF_y = -2 * r;
  int8_t x = 0;
  int8_t y = r;
  setpixel(x0, y0+r, color);
  setpixel(x0, y0-r, color);
  setpixel(x0+r, y0, color);
  setpixel(x0-r, y0, color);
  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;}
    x++;
    ddF_x += 2;
    f += ddF_x;
    setpixel(x0 + x, y0 + y, color);
    setpixel(x0 - x, y0 + y, color);
    setpixel(x0 + x, y0 - y, color);
    setpixel(x0 - x, y0 - y, color);
    setpixel(x0 + y, y0 + x, color);
    setpixel(x0 - y, y0 + x, color);
    setpixel(x0 + y, y0 - x, color);
    setpixel(x0 - y, y0 - x, color);}}

void NT7534_I2C::fillcircle(uint8_t x0, uint8_t y0, uint8_t r, uint8_t color) {
  int8_t f = 1 - r;
  int8_t ddF_x = 1;
  int8_t ddF_y = -2 * r;
  int8_t x = 0;
  int8_t y = r;
  for (uint8_t i=y0-r; i<=y0+r; i++) setpixel(x0, i, color);
  while (x<y) {
    if (f >= 0) {
      y--;
      ddF_y += 2;
      f += ddF_y;}
    x++;
    ddF_x += 2;
    f += ddF_x;
    for (uint8_t i=y0-y; i<=y0+y; i++) {
      setpixel(x0+x, i, color);
      setpixel(x0-x, i, color);}
    for (uint8_t i=y0-x; i<=y0+x; i++) {
      setpixel(x0+y, i, color);
      setpixel(x0-y, i, color);}}}

void  NT7534_I2C::printchar(uint8_t x, uint8_t line, uint16_t n) {
  for (uint8_t i =0; i<5; i++ ) {
     setbyte(x, line, pgm_read_byte(font+(n*5)+i));
     x++;
  }
}

// n for ascii number (1 = smily)
void NT7534_I2C::write(uint16_t n) {
  uint8_t x = gx;
  uint8_t line = gy;
  for (uint8_t i =0; i<5; i++ ) {
     setbyte(x, line, pgm_read_byte(font+(n*5)+i));
     x++;
     gx++;
  }
}



void  NT7534_I2C::clearchar(uint8_t x, uint8_t line, char c) {
  for (uint8_t i =0; i<5; i++ ) {
     setbyte(x, line, 0x00);
     x++;}}

void NT7534_I2C::drawstring(uint8_t x, uint8_t line, const char *c) {
  while (c[0] != 0) {
    printchar(x, line, c[0]);
    c++;
    x += 6; // 6 pixels wide
    if (x + 6 >= LCDWIDTH) {
      x = 0;    // ran out of this line
      line++;}
    if (line >= (LCDHEIGHT/8))
      return;}}        // ran out of space :(

void NT7534_I2C::clearstring(uint8_t x, uint8_t line, const char *c) {
  while (c[0] != 0) {
    clearchar(x, line, c[0]);
    c++;
    x += 6; // 6 pixels wide
    if (x + 6 >= LCDWIDTH) {
      x = 0;    // ran out of this line
      line++;}
    if (line >= (LCDHEIGHT/8))
      return;}}        // ran out of space :(

void NT7534_I2C::drawbitmap(uint8_t x, uint8_t y, const uint8_t *bitmap,
                        uint8_t w, uint8_t h, uint8_t color) {
  for (uint8_t j=0; j<h; j++) {
    for (uint8_t i=0; i<w; i++ ) {
      if (pgm_read_byte(bitmap + i + (j/8)*w) & (1<<(j%8))) {
	setpixel(x+i, y-1+j, color);}}}}

void NT7534_I2C::setCursor(uint8_t col, uint8_t row) {
  gx = col*6;
  gy = row; }

void NT7534_I2C::print(uint8_t x, uint8_t line, const char *c) {
  x = x*6;
  while (c[0] != 0) {
    printchar(x, line, c[0]);
    c++;
    x += 6; // 6 pixels wide
    if (x + 6 >= LCDWIDTH) {
      x = 0;    // ran out of this line
      line++;}
    if (line >= (LCDHEIGHT/8))
      return;}}        // ran out of space :(

void NT7534_I2C::print(uint8_t line, const char *c) {
  uint8_t x = 0;
  while (c[0] != 0) {
    printchar(x, line, c[0]);
    c++;
    x += 6; // 6 pixels wide
    if (x + 6 >= LCDWIDTH) {
      x = 0;    // ran out of this line
      line++;}
    if (line >= (LCDHEIGHT/8))
      return;}}        // ran out of space :(

void NT7534_I2C::print(const char *c) {
  uint8_t x = gx;
  uint8_t line = gy;
  while (c[0] != 0) {
    printchar(x, line, c[0]);
    c++;
    gx=gx+6;
    x += 6; // 6 pixels wide
    if (x + 6 >= LCDWIDTH) {
      x = 0;    // ran out of this line
      line++;
      gy++;}
    if (line >= (LCDHEIGHT/8))
      return;}}        // ran out of space :(

void NT7534_I2C::print(int i) {
  char *c;
  uint8_t x = gx;
  uint8_t line = gy;
  itoa(i,c,10);
  while (c[0] != 0) {
    printchar(x, line, c[0]);
    c++;
    gx=gx+6;
    x += 6; // 6 pixels wide
    if (x + 6 >= LCDWIDTH) {
      x = 0;    // ran out of this line
      line++;
      gy++;}
    if (line >= (LCDHEIGHT/8))
      return;}}        // ran out of space :(


/*void NT7534_I2C::clearchar(uint8_t x, uint8_t line, char c) {
  x = x*8;
  uint8_t line = y;
  for (uint8_t i =0; i<5; i++ ) {
     setbyte(x, line, 0x00);
     x++;}}*/
