// -*- C++ -*-
/*
* clever.cpp
* www.weigu.lu
*
* The clever software is based on the Open EVSE Firmware
* https://openevse.com/
*
* Copyright (c) 2011-2015 Sam C. Lin <lincomatic@gmail.com>
* Copyright (c) 2011-2014 Chris Howell <chris1howell@msn.com>
* timer code Copyright (c) 2013 Kevin L <goldserve1@hotmail.com>
* portions Copyright (c) 2014-2015 Nick Sayer <nsayer@kfu.com>
* portions Copyright (c) 2015 Craig Kirkpatrick
* portions Copyright (c) 2015 William McBrine
* portions Copyright (c) 2018 www.weigu.lu
*
*
*
* Revised    Ver By		Reason
* 13/09/17   0.11 weigu    DELAYTIMER raus cl_openevse.h 139
*                          ButtonMenu raus cl_openevse.h 120
*                          ino: 128-130,771-2015,2148-50,2177-2179
*                          oe.nenah: 120,126-131,137-143,166,715-991,1079-1082
*                          stringsh an strings.ccp
* 20/03/18  0.23 weigu     i2caddr.h raus (an clever.h)
*                          RTClib raus, Include RTCLib NeiroN JeeLabs
*                          changed Line 58 (RTC_DS1307 to DS3231) DS3231 g_RTC;
*                          line 175 code chunk read temp rtc del and replaced
*                          with function from new RTC library
*
* This file is part of Clever.
*
* Clever is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 3, or (at your option)
* any later version.
*
* Clever is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with Clever; see the file COPYING.  If not, write to the
* Free Software Foundation, Inc., 59 Temple Place - Suite 330,
* Boston, MA 02111-1307, USA.
*/

#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <pins_arduino.h>
#include <Wire.h>
#include <RTClib.h> // Arduino include RTCLib NeiroN JeeLabs (or platformio.ini)
#include "clever.h"
#include <SPI.h>
#include <MFRC522.h> // Arduino include MFRC522 (or platformio.ini)

// Global variables //////////////////////////////////////////////////////////b

int g_flagAllowCharging = 0; //weigu
char g_sTmp[TMP_BUF_SIZE];
OnboardDisplay g_OBD;

#ifdef RTC  // Instantiate RTC and Delay Timer - GoldServe
DS3231 g_RTC;
DateTime g_CurrTime;
#if defined(RAPI)
void SetRTC(uint8_t y,uint8_t m,uint8_t d,uint8_t h,uint8_t mn,uint8_t s) {
  g_RTC.adjust(DateTime(y,m,d,h,mn,s));
}
void GetRTC(char *buf) {
  g_CurrTime = g_RTC.now();
  sprintf(buf,"%02d/%02d/%02d     %02d:%02d:%02d",g_CurrTime.day(),g_CurrTime.month(),g_CurrTime.year()-2108,g_CurrTime.hour(),g_CurrTime.minute(),g_CurrTime.second());
}
#endif // RAPI
#endif // RTC
#ifdef DELAYTIMER
DelayTimer g_DelayTimer;
#endif // DELAYTIMER

#ifdef RFID
const uint8_t RST_PIN = A3;
const uint8_t SS_PIN = A4;
byte g_RFID_NUID[4] = {0xFF,0xFF,0xFF,0xFF};
MFRC522 g_rfid(SS_PIN, RST_PIN);   // Create MFRC522 instance
#endif // RFID

#ifdef TEMPERATURE_MONITORING
TempMonitor g_TempMonitor;
#endif // TEMPERATURE_MONITORING

#ifdef KWH_RECORDING
unsigned long g_WattHours_accumulated;
unsigned long g_WattSeconds;
#endif // KWH_RECORDING

#define g_sHHMMfmt "%02d:%02d"

// Global variables //////////////////////////////////////////////////////////e


// if digits > 0, zero pad w/ to # digits
// WARNING: This function uses the *end* of g_sTmp as its buffer
char *u2a(unsigned long x,int8_t digits)
{
  char *s = g_sTmp + sizeof(g_sTmp);

  *--s = 0;

  do {
    *--s = '0'+ x % 10;
    x /= 10;
    --digits;
  } while (x || digits > 0);

  return s;
}

void EvseReset();

// wdt_init turns off the watchdog timer after we use it
// to reboot
void wdt_init(void) __attribute__((naked)) __attribute__((section(".init3")));
void wdt_init(void)
{
  MCUSR = 0;
  wdt_disable();

  return;
}

// Temerature monitoring /////////////////////////////////////////////////////b

#ifdef TEMPERATURE_MONITORING
#ifdef TEMPERATURE_MONITORING_NY
void TempMonitor::LoadThresh()
{
  m_ambient_thresh = eeprom_read_word((uint16_t *)EOFS_THRESH_AMBIENT);
  if (m_ambient_thresh == 0xffff) {
    m_ambient_thresh = TEMPERATURE_AMBIENT_THROTTLE_DOWN;
  }
  m_ir_thresh = eeprom_read_word((uint16_t *)EOFS_THRESH_IR);
  if (m_ir_thresh == 0xffff) {
    m_ir_thresh = TEMPERATURE_INFRARED_THROTTLE_DOWN;
  }
}

void TempMonitor::SaveThresh()
{
  eeprom_write_word((uint16_t *)EOFS_THRESH_AMBIENT,m_ambient_thresh);
  eeprom_write_word((uint16_t *)EOFS_THRESH_IR,m_ir_thresh);
}
#endif // TEMPERATURE_MONITORING_NY

void TempMonitor::Init()
{
  m_Flags = 0;
  m_DS3231_temperature = 230;   // the DS3231 RTC has a built in temperature sensor

#ifdef TEMPERATURE_MONITORING_NY
  LoadThresh();
#endif
}

void TempMonitor::Read()
{
  unsigned long curms = millis();
  if ((curms - m_LastUpdate) >= TEMPMONITOR_UPDATE_INTERVAL) {

#ifdef RTC
  #define DS3231_ADDRESS 0x68
  // This code chunk below reads the DS3231 RTC's internal temperature sensor
  Wire.beginTransmission(DS3231_ADDRESS);
  Wire.write(uint8_t(0x0e));
  Wire.write( 0x20 );               // write bit 5 to initiate conversion of temperature
  Wire.endTransmission();

  Wire.beginTransmission(DS3231_ADDRESS);
  Wire.write(uint8_t(0x11));
  Wire.endTransmission();

  Wire.requestFrom(DS3231_ADDRESS, 2);
  m_DS3231_temperature = (((int16_t)Wire.read()) << 2) | (Wire.read() >> 6);
  if (m_DS3231_temperature == 0x3FF) {
    // assume chip not present
    m_DS3231_temperature = 0;  // If the DS3231 is not present then return 0
  }
  else {
    if (m_DS3231_temperature & 0x0200) m_DS3231_temperature |= 0xFE00; // sign extend negative number
    // convert from .25C to .1C
    m_DS3231_temperature = (m_DS3231_temperature * 10) / 4;
  }

//m_DS3231_temperature = g_RTC.getTemp(); //easier with new lib :) weigu
#endif // RTC

    m_LastUpdate = curms;
  }
}
#endif // TEMPERATURE_MONITORING
// Temerature monitoring /////////////////////////////////////////////////////e

// OnboardDisplay////////////////////////////////////////////////////////////////b
OnboardDisplay::OnboardDisplay()
  #ifdef CL_I2C_LCD
  : m_Lcd(7,6,5,4,3) // rst, bl, en, rw. di
  #endif
{}

void OnboardDisplay::Init() {
  WDT_RESET();
  m_bFlags = OBDF_MONO_BACKLIGHT;
  #ifdef CL_I2C_LCD
  LcdBegin(LCD_MAX_CHARS_PER_LINE, 8);
  m_Lcd.clear();
  #ifdef SHOWLOGO
  m_Lcd.displaylogo(1);
  delay(100);
  m_Lcd.clear();
  #endif // SHOWLOGO
  LcdPrint_P(0,1,PSTR("  CREATIVE-LAB.LU"));
  LcdPrint_P(0,3,PSTR(" EV Energy Reloader"));
  LcdPrint_P(0,5,PSTR("      clever"));
  LcdPrint_P(0,7,PSTR("Ver.: "));
  LcdPrint_P(VERSTR);
  #endif // CL_I2C_LCD
  delay(1000);
  m_Lcd.clear();
  WDT_RESET(); }

void OnboardDisplay::LcdPrint(int x,int y,const char *s) {
  m_Lcd.setCursor(x,y);
  m_Lcd.print(s);
}

void OnboardDisplay::LcdPrint(int x,int y,int i) {
  m_Lcd.setCursor(x,y);
  m_Lcd.print(i);
}

void OnboardDisplay::LcdPrintHex(int x,int y,byte i) {
  char HEXH = (i >> 4) + 0x30;
  if (HEXH > 0x39) HEXH +=7;
  char HEXL = (i & 0x0F) + 0x30;
  if (HEXL > 0x39) HEXL +=7;
  m_Lcd.setCursor(x,y);
  m_Lcd.write(HEXH); // write for char!!
  m_Lcd.setCursor(x+1,y);
  m_Lcd.write(HEXL);
}

void OnboardDisplay::LcdPrint_P(PGM_P s) {
  strncpy_P(m_strBuf,s,LCD_MAX_CHARS_PER_LINE);
  m_strBuf[LCD_MAX_CHARS_PER_LINE] = 0;
  m_Lcd.print(m_strBuf);
}

void OnboardDisplay::LcdPrint_P(int y,PGM_P s) {
  strncpy_P(m_strBuf,s,LCD_MAX_CHARS_PER_LINE);
  m_strBuf[LCD_MAX_CHARS_PER_LINE] = 0;
  LcdPrint(y,m_strBuf);
}

void OnboardDisplay::LcdPrint_P(int x,int y,PGM_P s) {
  strncpy_P(m_strBuf,s,LCD_MAX_CHARS_PER_LINE);
  m_strBuf[LCD_MAX_CHARS_PER_LINE] = 0;
  m_Lcd.setCursor(x,y);
  m_Lcd.print(m_strBuf);
}

void OnboardDisplay::LcdMsg_P(PGM_P l1,PGM_P l2) {
  LcdPrint_P(Msg1_y,l1);
  LcdPrint_P(Msg2_y,l2);
}

// print at (0,y), filling out the line with trailing spaces
void OnboardDisplay::LcdPrint(int y,const char *s) {
  m_Lcd.setCursor(0,y);
  uint8_t i,len = strlen(s);
  if (len > LCD_MAX_CHARS_PER_LINE)
    len = LCD_MAX_CHARS_PER_LINE;
  for (i=0;i < len;i++) {
    m_Lcd.write(s[i]);
   }
  for (i=len;i < LCD_MAX_CHARS_PER_LINE;i++) {
    m_Lcd.write(' ');
  }
}

void OnboardDisplay::LcdMsg(const char *l1,const char *l2) {
  LcdPrintBusID();
  LcdPrint(Msg1_y,l1);
  LcdPrint(Msg2_y,l2); }

void OnboardDisplay::LcdPrintBusID() {
  g_OBD.LcdPrint_P(ID_x,ID_y,PSTR("BusID: "));
  g_OBD.LcdPrint_P(CLCS_BUSSTR);
  g_OBD.LcdPrint_P(CLCS_IDSTR);
}

void OnboardDisplay::LcdClrLine(int y) {
  LcdPrint(y,"                        "); //24
}


void OnboardDisplay::Update(int8_t updmode)
{
  if (updateDisabled()) return;
  uint8_t curstate = g_EvseController.GetState();
  uint8_t svclvl = g_EvseController.GetCurSvcLevel();
  int currentcap = g_EvseController.GetCurrentCapacity();
  unsigned long curms = millis();
  if (g_EvseController.StateTransition() || (updmode != OBD_UPD_NORMAL)) {
    curms += 1000; // trigger periodic update code below
    //sprintf(g_sTmp,g_sRdyLAstr,(int)svclvl,currentcap);
    sprintf(g_sTmp,g_sRdyLAstr,currentcap);
    switch(curstate) {
    case EVSE_STATE_A: // not connected
      LcdPrintBusID();
      LcdPrint(Imax_x,Imax_y,g_sTmp);
      LcdPrint_P(Msg1_y,g_psState_A);
      LcdClrLine(Msg2_y);
      // Display Timer and Stop Icon - GoldServe
      #ifdef DELAYTIMER
      g_DelayTimer.PrintTimerIcon();
      #endif //#ifdef DELAYTIMER
      #ifdef KWH_RECORDING
      sprintf(g_sTmp,"%5luWh",(g_WattSeconds / 3600) );
      LcdPrint(Wh_x,Wh_y,g_sTmp);
      sprintf(g_sTmp,"%6lukWh",(g_WattHours_accumulated / 1000));  // display accumulated kWh
      LcdPrint(kWh_x,kWh_y,g_sTmp);
      #endif // KWH_RECORDING
      break;
    case EVSE_STATE_B: // connected/not charging
      LcdPrintBusID();
      LcdPrint(Imax_x,Imax_y,g_sTmp);
      LcdPrint_P(Msg1_y,g_psState_B);
      LcdClrLine(Msg2_y);
      #ifdef CHARGE_LIMIT
      if (g_EvseController.GetChargeLimit()) { LcdWrite(229); }// lightning
      #endif //CHARGE_LIMIT
      #ifdef TIME_LIMIT
      if (g_EvseController.GetTimeLimit()) { LcdWrite(226); }// clock
      #endif //TIME_LIMIT
      // Display Timer and Stop Icon - GoldServe
      #ifdef DELAYTIMER
      g_DelayTimer.PrintTimerIcon();
      #endif //#ifdef DELAYTIMER
      #ifdef KWH_RECORDING
      sprintf(g_sTmp,"%5luWh",(g_WattSeconds / 3600) );
      LcdPrint(Wh_x,Wh_y,g_sTmp);
      sprintf(g_sTmp,"%6lukWh",(g_WattHours_accumulated / 1000));  // Display accumulated kWh
      LcdPrint(kWh_x,kWh_y,g_sTmp);
      #endif // KWH_RECORDING
      break;
    case EVSE_STATE_C: // charging
      LcdPrintBusID();
      LcdPrint(Imax_x,Imax_y,g_sTmp);
      LcdPrint_P(Msg1_y,g_psState_C);
      LcdClrLine(Msg2_y);
      // Display Timer and Stop Icon - GoldServe
      #ifdef CHARGE_LIMIT
      if (g_EvseController.GetChargeLimit()) { LcdWrite(229); } // lightning
      #endif //CHARGE_LIMIT
      #ifdef TIME_LIMIT
      if (g_EvseController.GetTimeLimit()) { LcdWrite(226); }// clock
      #endif
      #ifdef DELAYTIMER
      g_DelayTimer.PrintTimerIcon();
      #endif //#ifdef DELAYTIMER
      break;
    case EVSE_STATE_D: // vent required
      LcdMsg_P(g_psEvseError,g_psVentReq);
      break;
    case EVSE_STATE_DIODE_CHK_FAILED:
      LcdMsg_P(g_psEvseError,g_psDiodeChkFailed);
      break;

    #ifdef TEMPERATURE_MONITORING
    case EVSE_STATE_OVER_TEMPERATURE:    // overtemp message in Red on the RGB LCD
      LcdMsg_P(g_psSvcReq,g_psTemperatureFault);  //  SERVICE REQUIRED     OVER TEMPERATURE
    #endif
      break;
    case EVSE_STATE_NO_GROUND:
      if (updmode == OBD_UPD_HARDFAULT) {LcdMsg_P(g_psEvseError,g_psNoGround);}
      else { LcdPrint_P(Msg1_y,g_psNoGround); } //// 2nd line will be updated below with auto retry count
      break;
    case EVSE_STATE_STUCK_RELAY:
      LcdMsg_P(updmode == OBD_UPD_HARDFAULT ? g_psSvcReq : g_psEvseError,g_psStuckRelay);
      break;
    case EVSE_STATE_DISABLED:
      LcdPrintBusID();
      LcdPrint(Imax_x,Imax_y,g_sTmp);
      LcdPrint_P(Msg1_y,g_psDisabled);
      LcdClrLine(Msg2_y);
      break;
    case EVSE_STATE_SLEEPING:
      LcdPrintBusID();
      LcdPrint(Imax_x,Imax_y,g_sTmp);
      LcdPrint_P(Msg1_y,g_psSleeping);
      LcdClrLine(Msg2_y);
      break;
    default:
      LcdClear();
    }
  }
  // put anything that needs to be updated periodically here
  // the code below will only run once per second
  if ((curms-m_LastUpdateMs) >= 1000) {
    m_LastUpdateMs = curms;
    if (!g_EvseController.InHardFault() && (curstate == EVSE_STATE_NO_GROUND)) {
      strcpy(g_sTmp,g_sRetryIn);
      int resetsec = (int)(g_EvseController.GetResetMs() / 1000ul);
      if (resetsec >= 0) {
        sprintf(g_sTmp+sizeof(g_sTmp)-6,g_sHHMMfmt,resetsec / 60,resetsec % 60);
        strcat(g_sTmp,g_sTmp+sizeof(g_sTmp)-6);
        LcdPrint(Msg2_y,g_sTmp);
      }
      return;
    }
    #ifdef RFID
    if (g_rfid.PICC_IsNewCardPresent() && g_rfid.PICC_ReadCardSerial()) {
      g_OBD.LcdPrint_P(RFID_x,RFID_y,PSTR("ID: "));
      for(byte i=0; i < 4; i++) { // UID size = 4 byte
          g_RFID_NUID[i] = g_rfid.uid.uidByte[i];
          g_OBD.LcdPrintHex(RFID_x+4+i*2,RFID_y,g_RFID_NUID[i]);
      }
      RapiSendRFID_NUID();
      g_rfid.PICC_HaltA(); // Halt PICC
      g_rfid.PCD_StopCrypto1(); // Stop encryption on PCD
    }
    #endif // RFID

    #ifdef RTC
    g_CurrTime = g_RTC.now();
    #endif
    #if defined(AMMETER)
    if (((curstate == EVSE_STATE_C) || g_EvseController.AmmeterCalEnabled()) && AmmeterIsDirty()) {
      SetAmmeterDirty(0);
      uint32_t current = g_EvseController.GetChargingCurrent();
      // display only if > 1000
      if (current >= 1000) {
        int a = current / 1000;
        int ma = (current % 1000) / 100;
        if (ma > 9) {
          ma = 0;
          a++;
        }
        sprintf(g_sTmp,"%3d.%dA",a,ma);
      }
      else { strcpy_P(g_sTmp,PSTR("    0A")); }
      LcdPrint(I_x,I_y,g_sTmp);
    }
    #endif // AMMETER
    if (curstate == EVSE_STATE_C) {
      #ifndef KWH_RECORDING
      time_t elapsedTime = g_EvseController.GetElapsedChargeTime();
      #endif
      #ifdef KWH_RECORDING
      uint32_t current = g_EvseController.GetChargingCurrent();
      g_WattSeconds =  g_WattSeconds + (VOLTS * current / 1000);
      sprintf(g_sTmp,"%5luWh",(g_WattSeconds / 3600) );
      LcdPrint(Wh_x,Wh_y,g_sTmp);
      sprintf(g_sTmp,"%6lukWh",(g_WattHours_accumulated / 1000));  // display accumulated kWh
      LcdPrint(kWh_x,kWh_y,g_sTmp);
      #endif // KWH_RECORDING
      #ifdef TEMPERATURE_MONITORING
      if ((g_TempMonitor.OverTemperature()) || TEMPERATURE_DISPLAY_ALWAYS)  {
        g_OBD.LcdClearLine(1);
        const char *tempfmt = "%2d.%1dC";
        #ifdef RTC
        if ( g_TempMonitor.m_DS3231_temperature != 0) {
        // it returns 0 if it is not present
          sprintf(g_sTmp,tempfmt,g_TempMonitor.m_DS3231_temperature/10, abs(g_TempMonitor.m_DS3231_temperature % 10));      //  sensor built into the DS3231 RTC Chip
          LcdPrint(5,1,g_sTmp);
        }
        if (g_TempMonitor.BlinkAlarm() && g_TempMonitor.OverTemperatureShutdown()){ // Blink Red off-and-on while over the temperature shutdown limit, zero current should flow to the EV
          g_TempMonitor.SetBlinkAlarm(0);  // toggle the alarm flag so we can blink
        }
        else {
          g_TempMonitor.SetBlinkAlarm(1);  // toggle the alarm flag so we can blink
	      }
      }  // (g_TempMonitor.OverTemperature()) || TEMPERATURE_DISPLAY_ALWAYS)
      else  {
        SetRedLed(0);          // restore the normal TEAL backlight in case it
      }
      if (!(g_TempMonitor.OverTemperature() || TEMPERATURE_DISPLAY_ALWAYS)) {
      #endif // TEMPERATURE_MONITORING
        #ifndef KWH_RECORDING
        int h = hour(elapsedTime);          // display the elapsed charge time
        int m = minute(elapsedTime);
        int s = second(elapsedTime);
        sprintf(g_sTmp,"%02d:%02d:%02d",h,m,s);
        #ifdef RTC
        g_sTmp[8]=' ';
        g_sTmp[9]=' ';
        g_sTmp[10]=' ';
        sprintf(g_sTmp+11,g_sHHMMfmt,(int)g_CurrTime.hour(),(int)g_CurrTime.minu  te());
        #endif //RTC
        LcdPrint(1,g_sTmp);
        #endif // KWH_RECORDING
    #ifdef TEMPERATURE_MONITORING
      }
    #endif // TEMPERATURE_MONITORING
    } // curstate == EVSE_STATE_C
    // Display a new stopped LCD screen with Delay Timers enabled - GoldServe
    #ifdef DELAYTIMER
    else if (curstate == EVSE_STATE_SLEEPING) {
      LcdSetCursor(0,0);
      g_DelayTimer.PrintTimerIcon();
      //     LcdPrint_P(g_DelayTimer.IsTimerEnabled() ? g_psWaiting : g_psSleeping);
      LcdPrint_P(g_psSleeping);
      sprintf(g_sTmp,"%02d:%02d:%02d",g_CurrTime.hour(),g_CurrTime.minute(),g_CurrTime.second());
      LcdPrint(0,1,g_sTmp);
      if (g_DelayTimer.IsTimerEnabled()){
        LcdSetCursor(9,0);
        LcdWrite(228); // arrow
        LcdWrite(226); // clock
        sprintf(g_sTmp,g_sHHMMfmt,g_DelayTimer.GetStartTimerHour(),g_DelayTimer.GetStartTimerMin());
        LcdPrint(11,0,g_sTmp);
        LcdSetCursor(9,1);
        LcdWrite(227); // cube
        LcdWrite(226); // clock
        sprintf(g_sTmp,g_sHHMMfmt,g_DelayTimer.GetStopTimerHour(),g_DelayTimer.GetStopTimerMin());
        LcdPrint(11,1,g_sTmp);
      }
      else {
        sprintf(g_sTmp,g_sRdyLAstr,(int)svclvl,currentcap);
        LcdPrint(10,0,g_sTmp);
      }
    }
    #endif // DELAYTIMER
#endif // LCD16X2
  }

// Display ///////////////////////////////////////////////////////////////////e

#ifdef FT_ENDURANCE
  LcdSetCursor(0,0);
  sprintf(g_sTmp,"%d %d",g_CycleCnt,(int)g_EvseController.ReadACPins());
  LcdPrint(g_sTmp);
#endif // FT_ENDURANCE
}

void ProcessInputs() { // ProcessInputs() ------------------------------------b
#ifdef RAPI
  RapiDoCmd();
#endif
#ifdef SERIALCLI
  g_CLI.getInput();
#endif // SERIALCLI
/*#ifdef BTN_MENU
  g_BtnHandler.ChkBtn();
#endif // BTN_MENU*/
#ifdef TEMPERATURE_MONITORING
  g_TempMonitor.Read();  //   update temperatures once per second
#endif
} // ProcessInputs() ---------------------------------------------------------e

void EvseReset() { // EvseReset() --------------------------------------------b
  Wire.begin();
#ifdef RTC
  g_RTC.begin();
#ifdef DELAYTIMER
  g_DelayTimer.Init();
#endif  // DELAYTIMER
#endif // RTC
  g_OBD.Init();
#ifdef RAPI
  RapiInit();
#endif
  g_EvseController.Init();
} // EvseReset() -------------------------------------------------------------e



void setup() { // Setup (begin) ++++++++++++++++++++++++++++++++++++++++++++++b
  wdt_disable();
  pinMode(A6,OUTPUT);   // PA6 to DE and /RE (Max 485)
  digitalWrite(A6,LOW); // Listen Receiver enabled
  delay(400);           // give I2C devices time to be ready before running
                        // code that wants to initialize I2C devices.
                        //  Otherwise a hang can occur upon powerup.
  Serial.begin(SERIAL_BAUD);


  #ifdef RFID
  SPI.begin();          // Init SPI bus
  g_rfid.PCD_Init();   // Init MFRC522 module
  delay(1000);
  g_rfid.PCD_DumpVersionToSerial();
  #endif // RFID

/*#ifdef BTN_MENU
  g_BtnHandler.init();
#endif // BTN_MENU*/
  EvseReset();
  #ifdef TEMPERATURE_MONITORING
  g_TempMonitor.Init();
  #endif // TEMPERATURE_MONITORING
  WDT_ENABLE();
  #ifdef KWH_RECORDING
  // Check for unitialized eeprom condition so it can begin at 0kWh
  // Set the four bytes to zero just once in the case of unitialized eeprom
  if (eeprom_read_dword((uint32_t*)EOFS_KWH_ACCUMULATED) == 0xffffffff) {
    eeprom_write_dword((uint32_t*)EOFS_KWH_ACCUMULATED,0);
    }
  // get the stored value for the kWh from eeprom
  g_WattHours_accumulated = eeprom_read_dword((uint32_t*)EOFS_KWH_ACCUMULATED);
  #endif // KWH_RECORDING
  g_EvseController.EnableDiodeCheck(1);
  g_EvseController.EnableGndChk(1);
} // Setup (end) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++e

int ix = 0; // for debugging
void loop() { // Mainloop (begin) ++++++++++++++++++++++++++++++++++++++++++++b

/*  #ifdef RFID
  if (g_rfid.PICC_IsNewCardPresent() && g_rfid.PICC_ReadCardSerial()) {
    g_OBD.LcdPrint_P(RFID_x,RFID_y,PSTR("ID: "));
    for(byte i=0; i < 4; i++) { // UID size = 4 byte
        g_RFID_NUID[i] = g_rfid.uid.uidByte[i];
        g_OBD.LcdPrintHex(RFID_x+4+i*2,RFID_y,g_RFID_NUID[i]);
    }
    RapiSendRFID_NUID();
    g_rfid.PICC_HaltA(); // Halt PICC
    g_rfid.PCD_StopCrypto1(); // Stop encryption on PCD
  }
  #endif // RFID*/


  // print time
  char RTCbuf[20]; //display clock
  GetRTC(RTCbuf);
  g_OBD.LcdPrint(RTC_x,RTC_y,RTCbuf);

  WDT_RESET();

  g_EvseController.Update();
  g_OBD.Update(); //update display
  ProcessInputs(); //rapi!

  // Delay Timer Handler - GoldServe
/*  #ifdef DELAYTIMER
  g_OBD.LcdPrint(16,6,ix+4);
  g_DelayTimer.CheckTime();
  #endif //#ifdef DELAYTIMER*/
} // Mainloop (end) ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++e
