// -*- C++ -*-
/*
 * clever.h
 * www.weigu.lu
 *
 * The clever software is based on the Open EVSE Firmware
 * https://openevse.com/
 *
 * This file is part of Clever.

 * Clever is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.

 * Clever is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with Clever; see the file COPYING.  If not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <pins_arduino.h>
#include <Wire.h>
#include "Time.h"
#include "avrstuff.h"

#define VERSION "V 0.29 31/05/18"
#define SHOWLOGO                       // show CL Logo at startup
#define CLCS_BUS '1'                   // EIA485 bus number (see RAPI)
#define CLCS_ID  '1'                   // number on bus
#define CLCS_BUS_STR "1"               // EIA485 bus number (see RAPI)
#define CLCS_ID_STR  "1"               // number on bus

//-- begin features ----------------------------------------------------------b

#define RFID // RFID
#define SHOW_DISABLED_TESTS // sdt before POST but only if display and Rapi!
#define RAPI // serial remote api
#define SERIAL_BAUD 250000 //weigu: reduce if necessary (problems if too low)
#define WATCHDOG // enable watchdog timer
#define TIME_LIMIT // charge for a specified amount of time and then stop
#define TEMPERATURE_MONITORING  // Temperature monitoring support
#define AMMETER // current measurement
// kWh Recording feature depends upon #AMMETER support
// comment out KWH_RECORDING to have the elapsed time and time of day displayed on the second line of the LCD
#ifdef AMMETER
#define KWH_RECORDING
#ifdef KWH_RECORDING
#define CHARGE_LIMIT // stop charging after a certain kWh reached
#endif // KWH_RECORDING
#endif //AMMETER
//Only one of both displays!
#define CL_I2C_LCD //Creative Labs recycled GTK281 128x64 (MCP23017)
//#define CL_I2C_OLED // Newheaven OLED (// to I2C with MCP23017)

// Definitions of display positions xy for display data
#define RTC_x 0 //define where to display clock on display
#define RTC_y 7
#define ID_x 0 //define where to display bus/ID on display
#define ID_y 0
#define RFID_x 0 //define where to display RFID NUID on display
#define RFID_y 6
#define Imax_x 13 //define where to display Imax on display
#define Imax_y 0
#define I_x 0 //define where to display Imax on display
#define I_y 3
#define Wh_x 0 //define where to display Wh on display
#define Wh_y 4
#define kWh_x 10 //define where to display kWh on display
#define kWh_y 4
#define Msg1_x 0 //define where to display Message 1
#define Msg1_y 2
#define Msg2_x 0 //define where to display Message 2
#define Msg2_y 3

#define ADVPWR // Advanced Powersupply... Ground check, stuck relay
// valid only if ADVPWR defined - for rectified MID400 chips which block
// half cycle (for ground check on both legs)
#define SAMPLE_ACPINS

// Option for RTC and DelayTime
// REQUIRES HARDWARE RTC: DS1307 or DS3231 connected via I2C
#define RTC // enable RTC & timer functions

#ifdef RTC
// Option for Delay Timer - GoldServe
//#define DELAYTIMER
#endif // RTC*/

// if defined, this pin goes HIGH when the EVSE is sleeping, and LOW otherwise
//#define SLEEP_STATUS_REG &PINB
//#define SLEEP_STATUS_IDX 4

// for stability testing - shorter timeout/higher retry count
//weigu 3/11/17
//#define RCD_TESTING

//-- end features ------------------------------------------------------------e
//-- begin functional tests --------------------------------------------------b

//#define NOCHECKS // for debugging ONLY - turns off all safety checks
// DO NOT USE FT_xxx. FOR FUNCTIONAL TESTING ONLY
//
// test delay between start of sleep mode and opening relay
// connect EV and put into charging state C
// relay should open after pilot goes back to state B or higher
// or after 3 sec, whichever happens first
// LCD will display SLEEP OPEN/THRESH if opened due to EV response
//              or  SLEEP OPEN/TIMEOUT if due to timeout
//
//#define FT_SLEEP_DELAY
//
// endurance test
// alternate state B 9 sec/state C 1 sec forever
// top line displays cycle count
// bottom line displays ac pin state
//#define FT_ENDURANCE


//#define FT_READ_AC_PINS // just read AC pins and display on line 1

//
//-- end functional tests  ---------------------------------------------------e

//-- begin configuration -----------------------------------------------------b

// WARNING: ALL DELAYS *MUST* BE SHORTER THAN THIS TIMER OR WE WILL GET INTO
// AN INFINITE RESET LOOP
#define WATCHDOG_TIMEOUT WDTO_4S
#define LCD_MAX_CHARS_PER_LINE 25
#define TMP_BUF_SIZE (LCD_MAX_CHARS_PER_LINE*2)


#define DEFAULT_CURRENT_CAPACITY 16 // current capacity in amps
#define MIN_CURRENT_CAPACITY 6 // minimum allowable current in amps
#define MAX_CURRENT_CAPACITY 32 // maximum current in amps (Clever 32A)

//CleverController
#define CURRENT_PIN 0 // analog current reading pin ADCx
#define PILOT_PIN 1 // analog pilot voltage reading pin ADCx

 // TEST PIN 1 for L1/L2, ground and stuck relay
#define ACLINE1_REG &PIND
#define ACLINE1_IDX 3
 // TEST PIN 2 for L1/L2, ground and stuck relay
#define ACLINE2_REG &PIND
#define ACLINE2_IDX 4

// digital Relay trigger pin
#define CHARGING_REG &PINB
#define CHARGING_IDX 0
// digital Relay trigger pin for second relay
#define CHARGING2_REG &PIND
#define CHARGING2_IDX 7
//digital Charging pin for AC relay
#define CHARGINGAC_REG &PINB
#define CHARGINGAC_IDX 1

// EEPROM offsets for settings
#define EOFS_CURRENT_CAPACITY_L1 0 // 1 byte
#define EOFS_CURRENT_CAPACITY_L2 1 // 1 byte
#define EOFS_FLAGS               2 // 2 bytes

// EEPROM offsets for Delay Timer function - GoldServe
#define EOFS_TIMER_FLAGS         4 // 1 byte
#define EOFS_TIMER_START_HOUR    5 // 1 byte
#define EOFS_TIMER_START_MIN     6 // 1 byte
#define EOFS_TIMER_STOP_HOUR     7 // 1 byte
#define EOFS_TIMER_STOP_MIN      8 // 1 byte

// AMMETER stuff
#define EOFS_CURRENT_SCALE_FACTOR 9 // 2 bytes
#define EOFS_AMMETER_CURR_OFFSET  11 // 2 bytes
#define EOFS_KWH_ACCUMULATED 13 // 4 bytes

// fault counters
#define EOFS_NOGND_TRIP_CNT    18 // 1 byte
#define EOFS_STUCK_RELAY_TRIP_CNT 19 // 1 byte

#define EOFS_VOLT_OFFSET 20 // 4 bytes
#define EOFS_VOLT_SCALE_FACTOR 24 // 2 bytes
#define EOFS_THRESH_AMBIENT 26 // 2 bytes
#define EOFS_THRESH_IR 28 // 2 bytes


// must stay within thresh for this time in ms before switching states
#define DELAY_STATE_TRANSITION 250
// must transition to state A from contacts closed in < 100ms according to spec
// but Leaf sometimes bounces from 3->1 so we will debounce it a little anyway
#define DELAY_STATE_TRANSITION_A 25

// for ADVPWR
#define GROUND_CHK_DELAY  1000 // delay after charging started to test, ms
#define STUCK_RELAY_DELAY 1000 // delay after charging opened to test, ms
#define RelaySettlingTime  250 // time for relay to settle in post, ms

// for SAMPLE_ACPINS - max number of ms to sample
#define AC_SAMPLE_MS 20 // 1 cycle @ 60Hz = 16.6667ms @ 50Hz = 20ms

#ifdef CL_I2C_LCD
#include "cl_i2c_nt7534_lcd_128x64.h"
#endif // CL_I2C_LCD

#ifdef CL_I2C_OLED
#include "cl_i2c_ssd1322_oled_128x64.h"
#endif // CL_I2C_OLED


/*#if defined(RGBLCD) || defined(I2CLCD)
// Using LiquidTWI2 for both types of I2C LCD's
// see http://blog.lincomatic.com/?p=956 for installation instructions
#ifdef RGBLCD
#define MCP23017 // Adafruit RGB LCD (PANELOLU2 is now supported without additional define)
#include "./LiquidTWI2.h"
#define LCD_I2C_ADDR 0x20 // for adafruit shield or backpack
#endif // RGBLCD
#endif // RGBLCD || I2CLCD*/

// button sensing pin
#define BTN_REG &PINC
#define BTN_IDX 3
#define BTN_PRESS_SHORT 50  // ms
#define BTN_PRESS_LONG 500 // ms
#define BTN_PRESS_VERYLONG 10000


#ifdef RTC
// Default start/stop timers for un-initialized EEPROMs.
// Makes it easy to compile in default time without need to set it up the first time.
#define DEFAULT_START_HOUR    0x00 //Start time: 00:05
#define DEFAULT_START_MIN     0x05
#define DEFAULT_STOP_HOUR     0x06 //End time: 6:55
#define DEFAULT_STOP_MIN      0x37
#endif // RTC

// for J1772.ReadPilot()
// 1x = 114us 20x = 2.3ms 100x = 11.3ms
#define PILOT_LOOP_CNT 100

#ifdef AMMETER
// This multiplier is the number of milliamps per A/d converter unit.

// First, you need to select the burden resistor for the CT. You choose the largest value possible such that
// the maximum peak-to-peak voltage for the current range is 5 volts. To obtain this value, divide the maximum
// outlet current by the Te. That value is the maximum CT current RMS. You must convert that to P-P, so multiply
// by 2*sqrt(2). Divide 5 by that value and select the next lower standard resistor value. For the reference
// design, Te is 1018 and the outlet maximum is 30. 5/((30/1018)*2*sqrt(2)) = 59.995, so a 56 ohm resistor
// is called for. Call this value Rb (burden resistor).

// Next, one must use Te and Rb to determine the volts-per-amp value. Note that the readCurrent()
// method calculates the RMS value before the scaling factor, so RMS need not be taken into account.
// (1 / Te) * Rb = Rb / Te = Volts per Amp. For the reference design, that's 55.009 mV.
// Each count of the A/d converter is 4.882 mV (5/1024). V/A divided by V/unit is unit/A. For the reference
// design, that's 11.26. But we want milliamps per unit, so divide that into 1000 to get 88.7625558. Round near...
//#define DEFAULT_CURRENT_SCALE_FACTOR 106 // for RB = 47 - recommended for 30A max
//#define DEFAULT_CURRENT_SCALE_FACTOR 184 // for RB = 27 - recommended for 50A max
// Craig K, I arrived at 213 by scaling my previous multiplier of 225 down by the ratio of my panel meter reading of 28 with the OpenEVSE uncalibrated reading of 29.6
// then upped the scale factor to 220 after fixing the zero offset by subtracing 900ma
//#define DEFAULT_CURRENT_SCALE_FACTOR 220 // for RB = 22 - measured by Craig on his new OpenEVSE V3
// NOTE: setting DEFAULT_CURRENT_SCALE_FACTOR TO 0 will disable the ammeter
// until it is overridden via RAPI
//#define DEFAULT_CURRENT_SCALE_FACTOR 220   // Craig K, average of three OpenEVSE controller calibrations

#define DEFAULT_CURRENT_SCALE_FACTOR 220   // OpenEVSE v2.5 and v3 with a 22 Ohm burden resistor (note that the schematic may say 28 Ohms by mistake)

// subtract this from ammeter current reading to correct zero offset

#define DEFAULT_AMMETER_CURRENT_OFFSET 0   // OpenEVSE v2.5 and v3 with a 22 Ohm burden resistor.  Could use a more thorough calibration exercise to nails this down.


#ifdef KWH_RECORDING
#define VOLTS 230 // Europe
#endif // KWH_RECORDING

// The maximum number of milliseconds to sample an ammeter pin in order to find three zero-crossings.
// one and a half cycles at 50 Hz is 30 ms.
#define CURRENT_SAMPLE_INTERVAL 35

// Once we detect a zero-crossing, we should not look for one for another quarter cycle or so. 1/4 // cycle at 50 Hz is 5 ms.
#define CURRENT_ZERO_DEBOUNCE_INTERVAL 5

#endif // AMMETER

#ifdef TEMPERATURE_MONITORING
#define TEMPERATURE_DISPLAY_ALWAYS 0     // Set this flag to 1 to always show temperatures on the bottom line of the 16X2 LCD
                                         // Set to it 0 to only display when temperatures become elevated
// #define TESTING_TEMPERATURE_OPERATION // Set this flag to play with very low sensor thresholds or to evaluate the code.
                                         // Leave it commented out instead to run with normal temperature thresholds.

// Temperature thresholds below are expressed as 520 meaning 52.0C to save from needing floating point library
// Keep any adjustments that you make at least 2C apart, giving things some hysterisis.  (example is 580 is 3C apart from 550)
// The RESTORE_AMPERAGE value must be lower than the THROTTLE_DOWN value
// The THROTTLE_DOWN value must be lower than the SHUTDOWN value
// The SHUTDOWN value must be lower than the PANIC value
#ifndef TESTING_TEMPERATURE_OPERATION
// normal oerational thresholds just below
// This is the temperature in the enclosure where we tell the car to draw 1/2 amperage.

#define TEMPERATURE_AMBIENT_THROTTLE_DOWN 650

// If the OpenEVSE responds nicely to the lower current drawn and temperatures in the enclosure
// recover to this level we can kick the current back up to the user's original amperage setting.

#define TEMPERATURE_AMBIENT_RESTORE_AMPERAGE 620

// This is the temperature in the enclosure where we tell the car to draw 1/4 amperage or 6A is minimum.

#define TEMPERATURE_AMBIENT_SHUTDOWN 680


//  At this temperature gracefully tell the EV to quit drawing any current, and leave the EVSE in
//  an over temperature error state.  The EVSE can be restart from the button or unplugged.
//  If temperatures get to this level it is advised to open the enclosure to look for trouble.

#define TEMPERATURE_AMBIENT_PANIC 710


#define TEMPERATURE_INFRARED_THROTTLE_DOWN 650    // This is the temperature seen  by the IR sensor where we tell the car to draw 1/2 amperage.
#define TEMPERATURE_INFRARED_RESTORE_AMPERAGE 600 // If the OpenEVSE responds nicely to the lower current drawn and temperatures in the enclosure
                                                  // recover to this level we can kick the current back up to the user's original amperage setting.
#define TEMPERATURE_INFRARED_SHUTDOWN 700         // This is the temperature in the enclosure where we tell the car to draw 1/4 amperage or 6A is minimum.

#define TEMPERATURE_INFRARED_PANIC 750            //  At this temperature gracefully tell the EV to quit drawing any current, and leave the EVSE in
                                                  //  an over temperature error state.  The EVSE can be restart from the button or unplugged.
                                                  //  If temperatures get to this level it is advised to open the enclosure to look for trouble.
#else  //TESTING_TEMPERATURE_OPERATION

// Below are good values for testing purposes at room temperature with an EV simulator and no actual high current flowing
#define TEMPERATURE_AMBIENT_THROTTLE_DOWN 290     // This is the temperature in the enclosure where we tell the car to draw 1/2 amperage.
#define TEMPERATURE_AMBIENT_RESTORE_AMPERAGE 270  // If the OpenEVSE responds nicely to the lower current drawn and temperatures in the enclosure
                                                  // recover to this level we can kick the current back up to the user's original amperage setting.
#define TEMPERATURE_AMBIENT_SHUTDOWN 310          // This is the temperature in the enclosure where we tell the car to draw 1/4 amperage or 6A is minimum.

#define TEMPERATURE_AMBIENT_PANIC 330             //  At this temperature gracefully tell the EV to quit drawing any current, and leave the EVSE in
                                                  //  an over temperature error state.  The EVSE can be restart from the button or unplugged.
                                                  //  If temperatures get to this level it is advised to open the enclosure to look for trouble.

#define TEMPERATURE_INFRARED_THROTTLE_DOWN 330    // This is the temperature seen  by the IR sensor where we tell the car to draw 1/2 amperage.
#define TEMPERATURE_INFRARED_RESTORE_AMPERAGE 270 // If the OpenEVSE responds nicely to the lower current drawn and temperatures in the enclosure
                                                  // recover to this level we can kick the current back up to the user's original amperage setting.
#define TEMPERATURE_INFRARED_SHUTDOWN 360         // This is the temperature in the enclosure where we tell the car to draw 1/4 amperage or 6A is minimum.

#define TEMPERATURE_INFRARED_PANIC 400            // At this temperature gracefully tell the EV to quit drawing any current, and leave the EVSE in
                                                  // an over temperature error state.  The EVSE can be restart from the button or unplugged.
                                                  // If temperatures get to this level it is advised to open the enclosure to look for trouble.

#endif // TESTING_TEMPERATURE_OPERATION

#endif // TEMPERATURE_MONITORING

// how long to show each disabled test on LCD
#define SHOW_DISABLED_DELAY 1500

//-- end configuration

typedef union union4b {
  int8_t i8;
  uint8_t u8;
  int16_t i16;
  uint16_t u16;
  int32_t i32;
  uint32_t u32;
  unsigned u;
  int i;
} UNION4B,*PUNION4B;


//-- begin class definitions

#ifdef WATCHDOG
#define WDT_RESET() wdt_reset() // pat the dog
#define WDT_ENABLE() wdt_enable(WATCHDOG_TIMEOUT)
#else
#define WDT_RESET()
#define WDT_ENABLE()
#endif // WATCHDOG

// OnboardDisplay.m_bFlags
#define OBDF_MONO_BACKLIGHT 0x01
#define OBDF_AMMETER_DIRTY  0x80
#define OBDF_UPDATE_DISABLED 0x40

// OnboardDisplay::Update()
#define OBD_UPD_NORMAL    0
#define OBD_UPD_FORCE     1 // update even if no state transition
#define OBD_UPD_HARDFAULT 2 // update w/ hard fault

//---class OnboardDisplay ----------------------------------------------------b
class OnboardDisplay {
  #ifdef RED_LED_REG
  DigitalPin pinRedLed;
  #endif
  #ifdef GREEN_LED_REG
  DigitalPin pinGreenLed;
  #endif

  #ifdef CL_I2C_LCD
  NT7534_I2C m_Lcd;
  #endif // CL_I2C_LCD

  uint8_t m_bFlags;
  char m_strBuf[LCD_MAX_CHARS_PER_LINE+1];
  unsigned long m_LastUpdateMs;
  int8_t updateDisabled() { return  m_bFlags & OBDF_UPDATE_DISABLED; }
  void MakeChar(uint8_t n, PGM_P bytes);

public:
  OnboardDisplay();
  void Init();

  void SetGreenLed(uint8_t state) {
  #ifdef GREEN_LED_REG
    pinGreenLed.write(state);
  #endif
  }

  void SetRedLed(uint8_t state) {
  #ifdef RED_LED_REG
  pinRedLed.write(state);
  #endif
  }

  #if defined(CL_I2C_OLED) || defined (CL_I2C_LCD)
  void LcdBegin(int x,int y) {
  #ifdef CL_I2C_LCD
    m_Lcd.begin2(x,y,0x19);
  #else
    m_Lcd.setMCPType(LTI_TYPE_MCP23017);
    m_Lcd.begin(x,y,2);
    m_Lcd.setBacklight(WHITE);
  #endif // CL_I2C_LCD
  }
  void LcdPrint(const char *s) {
    m_Lcd.print(s);
  }

  void LcdPrint_P(PGM_P s);
  void LcdPrint(int y,const char *s);
  void LcdPrint(int x,int y,int i);
  void LcdPrintHex(int x,int y,byte i);
  void LcdPrint_P(int y,PGM_P s);
  void LcdPrint(int x,int y,const char *s);
  void LcdPrint_P(int x,int y,PGM_P s);

  void LcdPrint(int i) {
    m_Lcd.print(i);
  }

  void LcdSetCursor(int x,int y) {
    m_Lcd.setCursor(x,y);
  }

  void LcdClearLine(int y) {
    m_Lcd.setCursor(0,y);
    for (uint8_t i=0;i < LCD_MAX_CHARS_PER_LINE;i++) {
      m_Lcd.write(' ');
    }
    m_Lcd.setCursor(0,y);
  }

  void LcdClear() {
    m_Lcd.clear();
  }

  void LcdWrite(uint16_t data) {
    m_Lcd.write(data);
  }

  void LcdMsg(const char *l1,const char *l2);
  void LcdMsg_P(PGM_P l1,PGM_P l2);
  void LcdPrintBusID();
  void LcdClrLine(int y);

  #endif // defined(CL_I2C_OLED) || defined (CL_I2C_LCD)

  #ifdef AMMETER
  void SetAmmeterDirty(uint8_t tf) {
    if (tf) m_bFlags |= OBDF_AMMETER_DIRTY;
    else m_bFlags &= ~OBDF_AMMETER_DIRTY;
  }
  int8_t AmmeterIsDirty() { return (m_bFlags & OBDF_AMMETER_DIRTY) ? 1 : 0; }
  #endif // AMMETER

  void DisableUpdate(int8_t on) {
    if (on) m_bFlags |= OBDF_UPDATE_DISABLED;
    else m_bFlags &= ~OBDF_UPDATE_DISABLED;
  }
  void Update(int8_t updmode=OBD_UPD_NORMAL); // OBD_UPD_xxx
};
//---class OnboardDisplay ----------------------------------------------------e

#ifdef RFID
extern byte g_RFID_NUID[4];
#endif // RFID

#ifdef RCD
#include "RCDev.h"
#endif // RCD

#define TEMPMONITOR_UPDATE_INTERVAL 1000ul
// TempMonitor.m_Flags
#define TMF_OVERTEMPERATURE          0x01
#define TMF_OVERTEMPERATURE_SHUTDOWN 0x02
#define TMF_BLINK_ALARM              0x04
class TempMonitor {
  uint8_t m_Flags;
  unsigned long m_LastUpdate;
public:
#ifdef TEMPERATURE_MONITORING_NY
  int16_t m_ambient_thresh;
  int16_t m_ir_thresh;
  int16_t m_TMP007_thresh;
#endif //TEMPERATURE_MONITORING_NY
  // these three temperatures need to be signed integers
  int16_t m_DS3231_temperature;   // the DS3231 RTC has a built in temperature sensor

  TempMonitor() {}
  void Init();
  void Read();

  void SetBlinkAlarm(int8_t tf) {
    if (tf) m_Flags |= TMF_BLINK_ALARM;
    else m_Flags &= ~TMF_BLINK_ALARM;
  }
  int8_t BlinkAlarm() { return (m_Flags & TMF_BLINK_ALARM) ? 1 : 0; }
  void SetOverTemperature(int8_t tf) {
    if (tf) m_Flags |= TMF_OVERTEMPERATURE;
    else m_Flags &= ~TMF_OVERTEMPERATURE;
  }
  int8_t OverTemperature() { return (m_Flags & TMF_OVERTEMPERATURE) ? 1 : 0; }
  void SetOverTemperatureShutdown(int8_t tf) {
    if (tf) m_Flags |= TMF_OVERTEMPERATURE_SHUTDOWN;
    else m_Flags &= ~TMF_OVERTEMPERATURE_SHUTDOWN;
  }
  int8_t OverTemperatureShutdown() { return (m_Flags & TMF_OVERTEMPERATURE_SHUTDOWN) ? 1 : 0; }
#ifdef TEMPERATURE_MONITORING_NY
  void LoadThresh();
  void SaveThresh();
#endif //TEMPERATURE_MONITORING_NY
};

#include "CPilot.h"
#include "CleverController.h"

#ifdef DELAYTIMER
// Start Delay Timer class definition - GoldServe
class DelayTimer {
  uint8_t m_DelayTimerEnabled;
  uint8_t m_StartTimerHour;
  uint8_t m_StartTimerMin;
  uint8_t m_StopTimerHour;
  uint8_t m_StopTimerMin;
  uint8_t m_CurrHour;
  uint8_t m_CurrMin;
  unsigned long m_LastCheck;
public:
  DelayTimer(){
    m_LastCheck = - (60ul * 1000ul);
  };
  void Init();
  void CheckTime();
  void Enable();
  void Disable();
  uint8_t IsTimerEnabled(){
    return m_DelayTimerEnabled;
  };
  uint8_t GetStartTimerHour(){
    return m_StartTimerHour;
  };
  uint8_t GetStartTimerMin(){
    return m_StartTimerMin;
  };
  uint8_t GetStopTimerHour(){
    return m_StopTimerHour;
  };
  uint8_t GetStopTimerMin(){
    return m_StopTimerMin;
  };
  void SetStartTimer(uint8_t hour, uint8_t min){
    m_StartTimerHour = hour;
    m_StartTimerMin = min;
    eeprom_write_byte((uint8_t*)EOFS_TIMER_START_HOUR, m_StartTimerHour);
    eeprom_write_byte((uint8_t*)EOFS_TIMER_START_MIN, m_StartTimerMin);
    g_EvseController.SaveSettings();
  };
  void SetStopTimer(uint8_t hour, uint8_t min){
    m_StopTimerHour = hour;
    m_StopTimerMin = min;
    eeprom_write_byte((uint8_t*)EOFS_TIMER_STOP_HOUR, m_StopTimerHour);
    eeprom_write_byte((uint8_t*)EOFS_TIMER_STOP_MIN, m_StopTimerMin);
    g_EvseController.SaveSettings();
  };
  uint8_t IsTimerValid(){
     if (m_StartTimerHour || m_StartTimerMin || m_StopTimerHour || m_StopTimerMin){ // Check not all equal 0
       if ((m_StartTimerHour == m_StopTimerHour) && (m_StartTimerMin == m_StopTimerMin)){ // Check start time not equal to stop time
         return 0;
       } else {
         return 1;
       }
     } else {
       return 0;
     }
  };
  void PrintTimerIcon();
};

#endif //ifdef DELAYTIMER
// -- end class definitions

char *u2a(unsigned long x,int8_t digits=0);
void ProcessInputs();
/*
extern char g_sPlus[];
extern char g_sSlash[];
extern char g_sColon[];
extern char g_sSpace[];
*/
#define g_sPlus "+"
#define g_sSlash "/"
#define g_sColon ":"
#define g_sSpace " "


#ifdef DELAYTIMER
extern DelayTimer g_DelayTimer;
#endif
/*#ifdef BTN_MENU
extern BtnHandler g_BtnHandler;
extern SettingsMenu g_SettingsMenu;
#endif // BTN_MENU*/
extern OnboardDisplay g_OBD;
extern char g_sTmp[TMP_BUF_SIZE];

extern int g_flagAllowCharging;


#ifdef KWH_RECORDING
extern unsigned long g_WattHours_accumulated;
extern unsigned long g_WattSeconds;
#endif // KWH_RECORDING
#ifdef TEMPERATURE_MONITORING
extern TempMonitor g_TempMonitor;
#endif // TEMPERATURE_MONITORING

#include "strings.h"
#include "rapi_proc.h"
